#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h> // read/write
#include <fcntl.h>   // file serial stuff 
#include <sys/ioctl.h>
#include "usb_i2c.h"

#define max(a,b) ((a)>(b)?(a):(b))
#define min(a,b) ((a)<(b)?(a):(b))

#define DEBUG 0

// This is a set of utility routines that help use the Devantech USB-I2C
// driver.  See the other programs that came with it for more examples.
// Included:
//  SerialOpen - opens serial device for writing.  Sets global mFD.
//  usb_i2c_cmd - send command to USB-I2C device itself, 
//                 return in MALLOCd
//                 buffer (NOTE: you must free this buffer when you're done!!!)
//  Read_i2c - reads register(s) of an I2C Device on the bus.
//                 return in MALLOCd
//                 buffer (NOTE: you must free this buffer when you're done!!!)
//  Write_i2c - writes a byte to a register of an I2C device on the bus.



struct termios mData;

int SerialOpen( const char* portDevStr, int portBaud ) {
	mFD = open(portDevStr,O_RDWR|O_NOCTTY);
  	if (mFD<0){
	  fprintf(stderr,"SerialOpen(): Could not open %s\n",portDevStr);
	  perror("SerialOpen()");
	  return -1;
	  }
#if DEBUG>5
	  else{
				 printf("Opened %s for read/write\n",portDevStr);
	  }
#endif

	if (ioctl(mFD, TCGETA, &mData)) { perror("SerialOpen:ioctl()"); return -1; }
    if(tcgetattr(mFD, &mData)<0) { perror("SerialOpen:tcgetattr"); return -1;}
    if (cfsetispeed(&mData, portBaud)<0) {perror("SerialOpen:cfsetispeed"); return -1;}
    mData.c_cflag |= (CLOCAL | CREAD);
    mData.c_lflag &=~(ICANON|ECHO|ISIG);
    mData.c_cflag &=~PARENB;
    mData.c_cflag &=~CSTOPB;
    mData.c_cflag &=~CSIZE;
    mData.c_cflag |= CS8;
    mData.c_iflag  = IGNBRK | IGNPAR | INLCR;
    tcflush(mFD, TCIFLUSH);
    tcsetattr(mFD, TCSANOW, &mData);
    return mFD;
}

// -----------------------------------------------------------------
// This routine read/write from/to the Devantech USB-I2C module.
//  (cf. Read_i2c & Write_i2c which pass through to I2C Devices)
//
BYTE *usb_i2c_cmd(enum usb_i2c_cmds command, BYTE data1, BYTE data2) {
		BYTE *sbuf =NULL;
		int ndata,nwritten, nread;

		// See "Technical Specification - USB-I2C USB to I2C Communications
		// Module".
		switch(command){
				  case USB_I2C_REVISION:
				  case NEW_ADDRESS:
				  case UNUSED: // or "battery" on CM02
				  case SETPINS:
				  case GETPINS: 	
							 			ndata=1;  break;
				  case SCAN1: 		ndata=6;  break;
				  case SCAN2: 		ndata=9;  break;
				  case SCAN3: 		ndata=12; break;
				  case SCAN4: 		ndata=15; break;
				  case SCAN6: 		ndata=21; break;
				  case SCAN8: 		ndata=27; break;
				  case SCAN12: 	ndata=39; break;
				  case SCAN16: 	ndata=51; break;
				  case GETAD: 		ndata=4;  break;
				  default:
							 return (BYTE *)NULL;
		}
		sbuf = (BYTE *)malloc(max(4,ndata)*sizeof(BYTE));
		if (sbuf==NULL) {
				  fprintf(stderr,"Error allocating %d bytes for sbuf\n",ndata);
				  perror("Read_usb_i2c()");
				  return (BYTE *)NULL;
		}

		sbuf[0] = USB_I2C_CMD;	
		sbuf[1] = command;
		sbuf[2] = data1;
		sbuf[3] = data2;
	   nwritten = write(mFD, sbuf, 4);
		if (nwritten!=4) {
				  fprintf(stderr,"Wrote %d (instead of 4) bytes...\n",nwritten);
				  free(sbuf);
				  return (BYTE *)NULL;
		} 
#if DEBUG>5
		else {
				  printf("Reading %i bytes from USB-I2C...",ndata);
		}
#endif

		nread = read(mFD, sbuf, ndata); // read data back
		if (nread!=ndata) {
				  fprintf(stderr,"Read %d (instead of %i) bytes...",nread,ndata);
				  fprintf(stderr,"attempting to read %d more ...",ndata-nread);
				  usleep(10*1000); // wait for 10ms, attempt to read the rest.
				  nread += read(mFD,sbuf+nread,ndata-nread);
				  if (nread!=ndata) {
				      fprintf(stderr,"Still couln't finish. Read %d (instead of %i) bytes...",nread,ndata);
						free(sbuf);
						return (BYTE *)NULL;
				  }
		} 
#if DEBUG>5
		else {
				  printf("done.\n");
		}
#endif

		// success
		return sbuf;
}


// -----------------------------------------------------------------
// These routines read/write from/to the Devantech USB-I2C module.
//  (cf. Read_i2c & Write_i2c which pass through to I2C Devices)
//
BYTE *Read_i2c(BYTE addr, BYTE start_reg, unsigned int ndata) {
		BYTE *sbuf;
		int nwritten, nread;

		sbuf = (BYTE *)malloc(max(ndata,4)*sizeof(BYTE));
		if (sbuf==NULL) {
				  fprintf(stderr,"Error allocating %d bytes for sbuf\n",ndata);
				  perror("Read_i2c()");
				  return (BYTE *)NULL;
		}

		sbuf[0] = I2C_CMD;	
		sbuf[1] = addr+READ;
		sbuf[2] = start_reg;
		sbuf[3] = ndata; // num data byte
	   nwritten = write(mFD, sbuf, 4);
		if (nwritten!=4) {
				  fprintf(stderr,"Wrote %d (instead of 4) bytes...\n",nwritten);
				  free(sbuf);
				  return (BYTE *)NULL;
		} 
#if DEBUG>5
		else {
				  printf("Reading %i bytes from address 0x%02x (and up)\n",ndata,addr);
		}
#endif

		nread = read(mFD, sbuf, ndata); // read data back
		if (nread!=ndata) {
#if DEBUG>7
				  fprintf(stderr,"Read %d (instead of %i) bytes...",nread,ndata);
				  fprintf(stderr,"attempting to read %d more ...",ndata-nread);
#endif
				  nread += read(mFD,sbuf+nread,ndata-nread);
				  if (nread!=ndata) {
				      fprintf(stderr,"Still couln't finish. Read %d (instead of %i) bytes...",nread,ndata);
						free(sbuf);
						return (BYTE *)NULL;
				  }
#if DEBUG>7
				  else {
							 fprintf(stderr,"Got em!\n");
				  }
#endif
		}

		// success
		return sbuf;
}

int Write_i2c(BYTE addr, BYTE reg, BYTE data) {
		BYTE sbuf[10];
		int nwritten, nread;

		sbuf[0] = I2C_CMD;	
		sbuf[1] = addr+WRITE;
		sbuf[2] = reg;
		sbuf[3] = 0x01; // 1 data byte
		sbuf[4] = data;
	   nwritten = write(mFD, &sbuf, 5);
		if (nwritten!=5) {
				  fprintf(stderr,"Wrote %d (instead of 5) bytes...\n",nwritten);
				  return -1;
		} 
#if DEBUG>5
		else {
				  printf("Wrote 0x%02x to address 0x%02x\n",data,addr+WRITE);
		}
#endif

		nread = read(mFD, sbuf, 1); // should be zero 
		if (nread!=1) {
				  fprintf(stderr,"Read %d (instead of 1) bytes...\n",nread);
				  return -1;
		}
		if (sbuf[0]==0) {
				  fprintf(stderr,"USB-I2C Returned Error.\n");
				  return -1;
		}

		// success
		return nwritten;
}

#ifdef INCMAIN
int main (void) {
	BYTE *sbuf;
	// Get USB-I2C firmware revision number
	sbuf = usb_i2c_cmd(USB_I2C_REVISION,0,0);
	printf("USB-I2C rev %i.\n",sbuf[0]);
	free(sbuf);
}
#endif
